"""Dirichlet.py

Maximum likelihood estimation and likelihood ratio tests of Dirichlet
distribution models of data.

Most of this package is a port of Thomas P. Minka's wonderful Fastfit MATLAB
code. Much thanks to him for that and his clear paper "Estimating a Dirichlet
distribution". See the following URL for more information:

    http://research.microsoft.com/en-us/um/people/minka/"""

import sys

import numpy as np
import scipy as sp
import scipy.stats as stats
from numpy import (
    arange,
    array,
    asanyarray,
    asarray,
    diag,
    exp,
    isscalar,
    log,
    ndarray,
    ones,
    vstack,
    zeros,
)
from numpy.linalg import norm
from scipy.special import gammaln, polygamma, psi

from . import simplex

MAXINT = sys.maxsize

__all__ = [
    "loglikelihood",
    "meanprecision",
    "mle",
    "pdf",
    "test",
]

euler = -1 * psi(1)  # Euler-Mascheroni constant


class NotConvergingError(Exception):
    """Error when a successive approximation method doesn't converge
    """
    pass


def test(D1, D2, method="meanprecision", maxiter=None):
    """Test for statistical difference between observed proportions.

    Parameters
    ----------
    D1 : (N1, K) shape array
    D2 : (N2, K) shape array
        Input observations. ``N1`` and ``N2`` are the number of observations,
        and ``K`` is the number of parameters for the Dirichlet distribution
        (i.e. the number of levels or categorical possibilities).
        Each cell is the proportion seen in that category for a particular
        observation. Rows of the matrices must add up to 1.
    method : string
        One of ``'fixedpoint'`` and ``'meanprecision'``, designates method by
        which to find MLE Dirichlet distribution. Default is
        ``'meanprecision'``, which is faster.
    maxiter : int
        Maximum number of iterations to take calculations. Default is
        ``sys.maxint``.

    Returns
    -------
    D : float
        Test statistic, which is ``-2 * log`` of likelihood ratios.
    p : float
        p-value of test.
    a0 : (K,) shape array
    a1 : (K,) shape array
    a2 : (K,) shape array
        MLE parameters for the Dirichlet distributions fit to 
        ``D1`` and ``D2`` together, ``D1``, and ``D2``, respectively."""

    N1, K1 = D1.shape
    N2, K2 = D2.shape
    if K1 != K2:
        raise ValueError("D1 and D2 must have the same number of columns")

    D0 = vstack((D1, D2))
    a0 = mle(D0, method=method, maxiter=maxiter)
    a1 = mle(D1, method=method, maxiter=maxiter)
    a2 = mle(D2, method=method, maxiter=maxiter)

    D = 2 * (loglikelihood(D1, a1) + loglikelihood(D2, a2) - loglikelihood(D0, a0))
    return (D, stats.chi2.sf(D, K1), a0, a1, a2)


def pdf(alphas):
    """Returns a Dirichlet PDF function

    Parameters
    ----------
    alphas : (K,) shape array
        The parameters for the distribution of shape ``(K,)``.

    Returns
    -------
    function
        The PDF function, takes an ``(N, K)`` shape input and gives an
        ``(N,)`` output.
    """
    alphap = alphas - 1
    c = np.exp(gammaln(alphas.sum()) - gammaln(alphas).sum())

    def dirichlet(xs):
        """Dirichlet PDF

        Parameters
        ----------
        xs : (N, K) shape array
            The ``(N, K)`` shape input matrix
        
        Returns
        -------
        (N,) shape array
            Point value for PDF
        """
        return c * (xs ** alphap).prod(axis=1)

    return dirichlet


def meanprecision(a):
    """Mean and precision of a Dirichlet distribution.

    Parameters
    ----------
    a : (K,) shape array
        Parameters of a Dirichlet distribution.

    Returns
    -------
    mean : (K,) shape array
        Means of the Dirichlet distribution. Values are in [0,1].
    precision : float
        Precision or concentration parameter of the Dirichlet distribution."""

    s = a.sum()
    m = a / s
    return (m, s)


def loglikelihood(D, a):
    """Compute log likelihood of Dirichlet distribution, i.e. log p(D|a).

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    a : (K,) shape array
        Parameters for the Dirichlet distribution.

    Returns
    -------
    logl : float
        The log likelihood of the Dirichlet distribution"""
    N, K = D.shape
    logp = log(D).mean(axis=0)
    return N * (gammaln(a.sum()) - gammaln(a).sum() + ((a - 1) * logp).sum())


def mle(D, tol=1e-7, method="meanprecision", maxiter=None):
    """Iteratively computes maximum likelihood Dirichlet distribution
    for an observed data set, i.e. a for which log p(D|a) is maximum.

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    method : string
        One of ``'fixedpoint'`` and ``'meanprecision'``, designates method by
        which to find MLE Dirichlet distribution. Default is
        ``'meanprecision'``, which is faster.
    maxiter : int
        Maximum number of iterations to take calculations. Default is
        ``sys.maxint``.

    Returns
    -------
    a : (K,) shape array
        Maximum likelihood parameters for Dirichlet distribution."""

    if method == "meanprecision":
        return _meanprecision(D, tol=tol, maxiter=maxiter)
    else:
        return _fixedpoint(D, tol=tol, maxiter=maxiter)


def _fixedpoint(D, tol=1e-7, maxiter=None):
    """Simple fixed point iteration method for MLE of Dirichlet distribution

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is
        ``sys.maxint``.

    Returns
    -------
    a : (K,) shape array
        Fixed-point estimated parameters for Dirichlet distribution."""
    logp = log(D).mean(axis=0)
    a0 = _init_a(D)

    # Start updating
    if maxiter is None:
        maxiter = MAXINT
    for i in range(maxiter):
        a1 = _ipsi(psi(a0.sum()) + logp)
        # Much faster convergence than with the more obvious condition
        # `norm(a1-a0) < tol`
        if abs(loglikelihood(D, a1) - loglikelihood(D, a0)) < tol:
            return a1
        a0 = a1
    raise NotConvergingError(
        "Failed to converge after {} iterations, values are {}.".format(maxiter, a1)
    )


def _meanprecision(D, tol=1e-7, maxiter=None):
    """Mean/precision method for MLE of Dirichlet distribution

    Uses alternating estimations of mean and precision.

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is
        ``sys.maxint``.

    Returns
    -------
    a : (K,) shape array
        Estimated parameters for Dirichlet distribution."""
    logp = log(D).mean(axis=0)
    a0 = _init_a(D)
    s0 = a0.sum()
    if s0 < 0:
        a0 = a0 / s0
        s0 = 1
    elif s0 == 0:
        a0 = ones(a0.shape) / len(a0)
        s0 = 1
    m0 = a0 / s0

    # Start updating
    if maxiter is None:
        maxiter = MAXINT
    for i in range(maxiter):
        a1 = _fit_s(D, a0, logp, tol=tol)
        s1 = sum(a1)
        a1 = _fit_m(D, a1, logp, tol=tol)
        m = a1 / s1
        # Much faster convergence than with the more obvious condition
        # `norm(a1-a0) < tol`
        if abs(loglikelihood(D, a1) - loglikelihood(D, a0)) < tol:
            return a1
        a0 = a1
    raise NotConvergingError(
        f"Failed to converge after {maxiter} iterations, " f"values are {a1}."
    )


def _fit_s(D, a0, logp, tol=1e-7, maxiter=1000):
    """Update parameters via MLE of precision with fixed mean

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    a0 : (K,) shape array
        Current parameters for Dirichlet distribution
    logp : (K,) shape array
        Mean of log-transformed D across N observations
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is 1000.

    Returns
    -------
    (K,) shape array
        Updated parameters for Dirichlet distribution."""
    s1 = a0.sum()
    m = a0 / s1
    mlogp = (m * logp).sum()
    for i in range(maxiter):
        s0 = s1
        g = psi(s1) - (m * psi(s1 * m)).sum() + mlogp
        h = _trigamma(s1) - ((m ** 2) * _trigamma(s1 * m)).sum()

        if g + s1 * h < 0:
            s1 = 1 / (1 / s0 + g / h / (s0 ** 2))
        if s1 <= 0:
            s1 = s0 * exp(-g / (s0 * h + g))  # Newton on log s
        if s1 <= 0:
            s1 = 1 / (1 / s0 + g / ((s0 ** 2) * h + 2 * s0 * g))  # Newton on 1/s
        if s1 <= 0:
            s1 = s0 - g / h  # Newton
        if s1 <= 0:
            raise NotConvergingError(f"Unable to update s from {s0}")

        a = s1 * m
        if abs(s1 - s0) < tol:
            return a

    raise NotConvergingError(f"Failed to converge after {maxiter} iterations, " f"s is {s1}")


def _fit_m(D, a0, logp, tol=1e-7, maxiter=1000):
    """Update parameters via MLE of mean with fixed precision s

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.
    a0 : (K,) shape array
        Current parameters for Dirichlet distribution
    logp : (K,) shape array
        Mean of log-transformed D across N observations
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is 1000.

    Returns
    -------
    (K,) shape array
        Updated parameters for Dirichlet distribution."""
    s = a0.sum()
    for i in range(maxiter):
        m = a0 / s
        a1 = _ipsi(logp + (m * (psi(a0) - logp)).sum())
        a1 = a1 / a1.sum() * s

        if norm(a1 - a0) < tol:
            return a1
        a0 = a1

    raise NotConvergingError(f"Failed to converge after {maxiter} iterations, " f"s is {s}")


def _init_a(D):
    """Initial guess for Dirichlet alpha parameters given data D

    Parameters
    ----------
    D : (N, K) shape array
        ``N`` is the number of observations, ``K`` is the number of
        parameters for the Dirichlet distribution.

    Returns
    -------
    (K,) shape array
        Crude guess for parameters of Dirichlet distribution."""
    E = D.mean(axis=0)
    E2 = (D ** 2).mean(axis=0)
    return ((E[0] - E2[0]) / (E2[0] - E[0] ** 2)) * E


def _ipsi(y, tol=1.48e-9, maxiter=10):
    """Inverse of psi (digamma) using Newton's method. For the purposes
    of Dirichlet MLE, since the parameters a[i] must always
    satisfy a > 0, we define ipsi :: R -> (0,inf).
    
    Parameters
    ----------
    y : (K,) shape array
        y-values of psi(x)
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is 10.

    Returns
    -------
    (K,) shape array
        Approximate x for psi(x)."""
    y = asanyarray(y, dtype="float")
    x0 = np.piecewise(
        y,
        [y >= -2.22, y < -2.22],
        [(lambda x: exp(x) + 0.5), (lambda x: -1 / (x + euler))],
    )
    for i in range(maxiter):
        x1 = x0 - (psi(x0) - y) / _trigamma(x0)
        if norm(x1 - x0) < tol:
            return x1
        x0 = x1
    raise NotConvergingError(f"Failed to converge after {maxiter} iterations, " f"value is {x1}")


def _trigamma(x):
    return polygamma(1, x)



#!/usr/bin/env python

import sys
from setuptools import setup

# Note, any changes here should be replicated in the `tox.ini` file

if not sys.version_info[:2] >= (3, 6):
    raise Exception("Only supports Python versions >= 3.6")
install_requires = [
    "scipy >= 1.4.1",
    "numpy >= 1.18.1",
    "matplotlib >= 3.2.0",
]

setup(
    name="dirichlet",
    version="0.9",
    description="Calculates Dirichlet test and plots 2-simplex Dirichlets",
    author="Eric Suh",
    author_email="contact@ericsuh.com",
    packages=["dirichlet"],
    install_requires=install_requires,
    url="http://github.com/ericsuh/dirichlet",
    download_url="https://github.com/ericsuh/dirichlet/zipball/master",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    long_description="""\
Dirichlet Test
--------------

Calculates nested model Dirichlet test of independence by finding maximum
likelihood estimates of Dirichlet distributions for different data sets
and comparing to the null hypothesis of the data being derived from one
distribution.

In addition, ``dirichlet.simplex`` module creates scatter, contour, and filled
contour 2-simplex plots.
""",
)



"""Tests for the functions in the Dirichlet package."""

import sys

import numpy
import pytest
from numpy.linalg import norm
from scipy.special import psi

import dirichlet

TOL = 1.48e-8


def test_ipsi():
    x_input = numpy.logspace(-5, 5)
    assert dirichlet.dirichlet._ipsi(psi(0.01)) - 0.01 < TOL
    assert norm(dirichlet.dirichlet._ipsi(psi(x_input)) - x_input) < TOL * len(x_input)


class TestDirichlet:
    # seed must be convertible to 32 bit unsigned integers
    numpy.random.seed(12987529)
    a0 = numpy.array([100, 299, 100])
    D0 = numpy.random.dirichlet(a0, 1000)
    logl0 = dirichlet.loglikelihood(D0, a0)

    D0a = numpy.random.dirichlet(a0, 1000)
    logl0a = dirichlet.loglikelihood(D0a, a0)

    a1 = numpy.array([50, 50, 90])
    D1 = numpy.random.dirichlet(a1, 1000)
    logl1 = dirichlet.loglikelihood(D1, a1)

    @pytest.mark.parametrize("method", ["fixedpoint", "meanprecision"])
    def test_mle(self, method):
        a0_fit = dirichlet.mle(self.D0, method=method)
        logl0_fit = dirichlet.loglikelihood(self.D0, a0_fit)
        assert norm(self.a0 - a0_fit) / norm(self.a0) < 0.1
        assert abs((logl0_fit - self.logl0) / logl0_fit) < 0.01

    @pytest.mark.parametrize("method", ["fixedpoint", "meanprecision"])
    def test_lltest(self, method):
        D, p, a0_fit, a1_fit, a2_fit = dirichlet.test(self.D0, self.D0a, method=method)
        # assert(p > 0.05) Need to do a uniform KS-test for uniform p-values
        assert norm(self.a0 - a0_fit) / norm(self.a0) < 0.1

        D, p, a0_fit, a1_fit, a2_fit = dirichlet.test(self.D0, self.D1, method=method)
        assert p < 0.05
        assert norm(self.a0 - a1_fit) / norm(self.a0) < 0.1
        assert norm(self.a1 - a2_fit) / norm(self.a1) < 0.1



import matplotlib
import matplotlib.pyplot as plt
import numpy as np

__all__ = [
    "cartesian",
    "barycentric",
    "contour",
    "contourf",
    "scatter",
]


def cartesian(points):
    """Converts array of barycentric coordinates on a 2-simplex to an array of
    Cartesian coordinates on a 2D triangle in the first quadrant.

    Parameters
    ----------
    points : (N, 3) shape array
        Points on a 2-simplex.

    Returns
    -------
    (N, 2) shape array
        Cartesian coordinate points.

    Examples
    --------
    >>> cartesian((1,0,0))
    array([0, 0])

    >>> cartesian((0,1,0))
    array([0, 1])

    >>> cartesian((0,0,1))
    array([0.5, 0.8660254037844386]) # == [0.5, sqrt(3)/2]"""
    points = np.asanyarray(points)
    ndim = points.ndim  # will use this to have similar output shape to input
    if ndim == 1:
        points = points.reshape((1, points.size))
    d = points.sum(axis=1)  # in case values aren't normalized
    x = 0.5 * (2 * points[:, 1] + points[:, 2]) / d
    y = (np.sqrt(3.0) / 2) * points[:, 2] / d
    out = np.vstack([x, y]).T
    if ndim == 1:
        return out.reshape((2,))
    return out


def barycentric(points):
    """Inverse of :func:`cartesian`."""
    points = np.asanyarray(points)
    ndim = points.ndim
    if ndim == 1:
        points = points.reshape((1, points.size))
    c = (2 / np.sqrt(3.0)) * points[:, 1]
    b = (2 * points[:, 0] - c) / 2.0
    a = 1.0 - c - b
    out = np.vstack([a, b, c]).T
    if ndim == 1:
        return out.reshape((3,))
    return out


def scatter(points, vertexlabels=None, **kwargs):
    """Scatter plot of barycentric 2-simplex points on a 2D triangle.

    Parameters
    ----------
    points : (N, 3) shape array
        N points on a 2-simplex
    vertexlabels : (str, str, str)
        Labels for corners of the plot, in the order
        ``(a, b, c)`` where ``a == (1,0,0)``, ``b == (0,1,0)``,
        ``c == (0,0,1)``.
    **kwargs : any
        Arguments to :func:`plt.scatter`.
    
    Returns
    -------
    matplotlib.figure.Figure
        The drawn simplex plot figure."""
    if vertexlabels is None:
        vertexlabels = ("1", "2", "3")

    projected = cartesian(points)
    plt.scatter(projected[:, 0], projected[:, 1], **kwargs)

    _draw_axes(vertexlabels)
    return plt.gcf()


def contour(f, vertexlabels=None, **kwargs):
    """Contour line plot on a 2D triangle of a function evaluated at
    barycentric 2-simplex points.

    Parameters
    ----------
    f : function
        Function to evaluate on (N, 3) ndarray of coordinates
    vertexlabels : (str, str, str)
        Labels for corners of the plot, in the order
        ``(a, b, c)`` where ``a == (1,0,0)``, ``b == (0,1,0)``,
        ``c == (0,0,1)``.
    **kwargs : any
        Arguments to :func:`plt.tricontour`.
        
    Returns
    -------
    matplotlib.figure.Figure
        The drawn contour line plot figure."""
    return _contour(f, vertexlabels, contourfunc=plt.tricontour, **kwargs)


def contourf(f, vertexlabels=None, **kwargs):
    """Filled contour plot on a 2D triangle of a function evaluated at
    barycentric 2-simplex points.

    Function signature is identical to :func:`contour` with the caveat that
    ``**kwargs`` are passed on to :func:`plt.tricontourf`."""
    return _contour(f, vertexlabels, contourfunc=plt.tricontourf, **kwargs)


def _contour(f, vertexlabels=None, contourfunc=None, **kwargs):
    """Workhorse function for ``contour`` and ``contourf``.

    Parameters
    ----------
    contourfunc : function
        The contour plotting function to use for actual plotting."""

    if contourfunc is None:
        contourfunc = plt.tricontour
    if vertexlabels is None:
        vertexlabels = ("1", "2", "3")
    x = np.linspace(0, 1, 100)
    y = np.linspace(0, np.sqrt(3.0) / 2.0, 100)
    points2d = np.transpose([np.tile(x, len(y)), np.repeat(y, len(x))])
    points3d = barycentric(points2d)
    valid = (points3d.sum(axis=1) == 1.0) & ((0.0 <= points3d).all(axis=1))
    points2d = points2d[np.where(valid), :][0]
    points3d = points3d[np.where(valid), :][0]
    z = f(points3d)
    contourfunc(points2d[:, 0], points2d[:, 1], z, **kwargs)
    _draw_axes(vertexlabels)
    return plt.gcf()


def _draw_axes(vertexlabels):
    l1 = matplotlib.lines.Line2D([0, 0.5, 1.0, 0], [0, np.sqrt(3) / 2, 0, 0], color="k")
    axes = plt.gca()
    axes.add_line(l1)
    axes.xaxis.set_major_locator(matplotlib.ticker.NullLocator())
    axes.yaxis.set_major_locator(matplotlib.ticker.NullLocator())
    axes.text(-0.05, -0.05, vertexlabels[0])
    axes.text(1.05, -0.05, vertexlabels[1])
    axes.text(0.5, np.sqrt(3) / 2 + 0.05, vertexlabels[2])
    axes.set_xlim(-0.2, 1.2)
    axes.set_ylim(-0.2, 1.2)
    axes.set_aspect("equal")
    return axes



"""Tests for the functions in the Simplex module."""

import sys

import numpy as np
import numpy.linalg

from dirichlet import simplex

TOL = 1.48e-8


def assert_almost(x, y, tol=1.48e-8):
    x = np.asanyarray(x)
    y = np.asanyarray(y)
    assert numpy.linalg.norm(x - y) < tol


def test_cartesian():
    assert_almost(simplex.cartesian(np.array([1, 0, 0])), np.array([0, 0]))
    assert_almost(simplex.cartesian(np.array([0, 1, 0])), np.array([1, 0]))
    assert_almost(
        simplex.cartesian(np.array([0, 0, 1])), np.array([0.5, 0.5 * np.sqrt(3)])
    )
    assert_almost(simplex.cartesian(np.array([0.5, 0.5, 0])), np.array([0.5, 0]))
    assert_almost(
        simplex.cartesian(np.array([0, 0.5, 0.5])), np.array([0.75, np.sqrt(3) * 0.25])
    )


def test_simplex():
    assert_almost(simplex.barycentric(np.array([0, 0])), np.array([1, 0, 0]))
    assert_almost(simplex.barycentric(np.array([1, 0])), np.array([0, 1, 0]))
    assert_almost(
        simplex.barycentric(np.array([0.5, 0.5 * np.sqrt(3)])), np.array([0, 0, 1])
    )
